/**********************************************************************

	Command:  PASTEXT -- Paste Texts from clipboard

	Programmed by Kang, Hsin-Min.
	Release at Jun 28, 1997
	Revised at Feb 14, 1998 -- Replace "Run DDSTYLE" to DoCommand()
				   such that LiteMode can be serviced.
	Revised at Apr 22, 1998 Include Japenese Text in SJIS codes

	Copyright (c) 1997, TCAM Development House, Taipei
	All rights Reserved.

	Status: Standard Command, with full support.

	This program will insert selected Text lines read from clipboard
	buffer if they are available.

	This program requres DDSTYLE be present in the path where it
	can be reached by TwinCAD.

***********************************************************************/
//
#define CF_TEXT 	1
#define TRUE	1
#define FALSE	0
//
//	There had been a bug in some TCL runtime functions that can not
//	properly handle a string array of which the size is larger than
//	32KB (a typical signed/unsigned integer index problem). So, to
//	avoid possible failures on some earlier versions, the string
//	array size must be less than 32767 bytes. However, this should
//	be enough for most of the application cases.
//
#define LINESIZE	80
#define MAXLINE 	400
//
char	szTextBuf[MAXLINE][LINESIZE];	// Clipboard Text buffer
//
int	IsWin=1;
int	nMaxLine;		// Number of lines
int	nStart=-1;		// Starting line
int	nEnd=-1;		// Ending line
//
char	szText[15][80]; // Our display text buffer, see SetMsgText()
			// at the end of this file

COMMAND PasText()
{
	int	cmdecho, hasjcode;
	char	szBuf[128];

	switch(TRUE) {	// Test if we are in TwinCAD ...
	  case !identify("os_type"):
	  case (os_type()<0):	// Must be in DOS version!
		printf("\nPASTEXT requires TwinCAD.");
		exit(1);
	  case !identify("cb_query"):
		printf("\nSorry, you need to update your version of TwinCAD.");
		exit(1);
	}

	SetMsgText();	// Set up our display text

	memset(szTextBuf,0);  // Clear the data
	if ( cb_getdata(CF_TEXT,szTextBuf)<= 0 ) {
		printf(szText[0]/*"\nNo text available from clipboard."*/);
		exit(1);
	}
	//
	// Well, since cb_getdata() returns the total size of the data,
	// instead of the number of text lines available, we have to
	// count it by ourself. Also, since the empty line will be returned
	// as a null line, we must count it backward.
	//
	for(nMaxLine=MAXLINE;nMaxLine&&(szTextBuf[nMaxLine-1]=="");nMaxLine--);

	if ( nMaxLine == 0 ) {
		printf(szText[1]/*"\nNo effective text available from clipboard." */);
		exit(1);
	}
	if ( !MarkText() )	// User quit!
		exit(1);
	if ( nStart == -1 )	// Redundent, just to make sure again
		exit(1);
	if ( nStart > nEnd )	// Make sure the direction
		swap(&nStart,&nEnd);
	//
	// Now, we shall call DDSTYLE to setup text style control for the
	// text to create. A simple program interface is deviced using
	// the exbuf_io() function. See DDSTYLE.TCL for details.
	//
	szBuf = "%%DDSTYLE -INS";       // Tell DDSTYLE to Disable [Insert]
	exbuf_io(0,szBuf,sizeof(szBuf));

	cmdecho = @cmdecho;
	@cmdecho = 0;
	DoCommand("DDSTYLE");         // Go for it!
	@cmdecho = cmdecho;

	exbuf_io(1,szBuf,sizeof(szBuf));	// Get return message
	if ( szBuf == "##DDSTYLE CANCEL" )
		exit(1);	// User Quit!
	hasjcode = 0;
	if ( left$(szBuf,12) == "##DDSTYLE OK" ) {
		szBuf -= "##DDSTYLE OK";
		szBuf -= " ";
		if ( szBuf != "" )
			hasjcode = 1;
	}
	//
	// If DDSTYLE is not present, or it is not updated for the
	// new program interface, the program execution will always
	// reach here with hasjcode set to 0.
	//
	p1 = getp(szText[2]/*"\nText Insert Point: "*/);
	v1 = getangle(szText[3]/*"\nText Insert Angle <0.>: "*/,0,p1);
	cmdecho = @cmdecho;
	@cmdecho = 0;
	if ( hasjcode )
		command("text J "+szBuf+" p1 ^Mv1 @!\@szTextBuf[\@nStart]^M^M");
	else
		command("text p1 ^Mv1 @!\@szTextBuf[\@nStart]^M^M");
	while(++nStart<=nEnd)
		command("text ^M@!\@szTextBuf[\@nStart]^M^M");
	@cmdecho = cmdecho;
	exit(0);
}
//
//	Execute a command (internal, extended or external)
//
//	TwinCAD Lite can not execute RUN command!
//
int	DoCommand(char cmdname[])
{
	if ( identify(cmdname) )	// Is command exist?
		return command(cmdname);    // Yes, do it!
	if ( identify("SystemMode") ) {
		if ( SystemMode() >= 0 )   // Is it non-lite mode?
			return command("RUN "+cmdname);  // Yes, run the name file
		printf("\nCan not execute %s command!",cmdname);
		return 0;
	}
	return command("RUN "+cmdname); // Execute it anyway!
}
//
//	MarkText() -- Prompt for user to view/edit path
//
int	nTatrb=0x270;
int	nSatrb=0x270;
//
int	MarkText()
{
	int	MainWnd[]={3,24,0,78,0x370};
	int	nSubWnd[]={0x0801,0x0810,3,75,0x21f0};
	int	rSub[]={0,8,0,21,0x21f0};
	int	widMain;	// Main window handle
	int	widSub; 	// Sub window handle
	int	bSBar;		// Base ID of Sub Scroll Bar
	int	bSBarFirst;
	int	bSBarLast;
	int	which;		// Temporary
	int	sCurrent;
	int	sTotal;
	int	sUpd;
	int	sPageSize;
	int	i, j, k;
	int	rBase;
	int	bMarkAll;
	int	bUnMark;
	int	bCancel;
	int	bOK;
	int	bStart;
	int	bEnd;
	int	nHasStateFnc;

	widMain = wndopen(MainWnd,
	     prints(szText[4]/*"Insert Texts from Clipboard [%d lines read]"*/,nMaxLine),
	     0x7f1e,0x8000);
	rSub[0] = 0;
	rSub[1] = nSubWnd[1]-nSubWnd[0];
	rSub[2] = 0;
	rSub[3] = nSubWnd[3]-nSubWnd[2];

	WndGroup(nSubWnd[0]-1,nSubWnd[2],(nSubWnd[3]-nSubWnd[2])+1,
		(nSubWnd[1]&0xff)-(nSubWnd[0]&0xff)+1,
		szText[5]/*"Please Mark the Text to Insert"*/,nSatrb);
	sTotal = nMaxLine;
	sCurrent = 0;
	widSub = wndopensub(nSubWnd,0);
	sPageSize = wnd_runit(0);
	j = sTotal-sPageSize;
	if ( j <= 0 )
		j = 1;
	bSBar = wnd_vscroll(0,0,0x1430,j,sCurrent);
	for(i=0;i<sPageSize;i++) {
		wnd_loct(i,0);
		bSBarLast = wnd_field(0,0xf0,(nSubWnd[3]-nSubWnd[2])-1);
		if ( i == 0 )
			bSBarFirst = bSBarLast;
	}
	rBase = nSubWnd[1]+1+0x401;
	wnd_select(widMain);
	bOK	= WndButton(rBase,4, 0x108,szText[6]/*"&OK"*/);
	bCancel = WndButton(rBase,16,0x108,szText[7]/*"&Cancel"*/);
	bMarkAll= WndButton(rBase,28,0x108,szText[8]/*"&All"*/);
	bUnMark = WndButton(rBase,40,0x108,szText[9]/*"&UnSelect"*/);
	bStart	= WndEntry(rBase,50,szText[10]/*"Range: "*/,0x104,0);
	bEnd	= WndEntry(rBase,55+len(szText[10]),"- ",0x104,0);
	nStart = nEnd = -1;
	if ( nHasStateFnc=identify("wnd_state") ) {
		wnd_state(bOK,1,1);
		wnd_state(bUnMark,1,1);
	}
	sUpd = 2;
	for(;;) {
	    if ( sUpd ) {
		j = sTotal-sPageSize;
		if ( j <= 0 )
		    j = 1;
		wnd_vscroll(1,bSBar,0x1430,j,sCurrent);
		if ( sUpd == 2 ) {
		    wnd_select(widSub);
		    for(i=0;i<sPageSize;i++) {
			wnd_loct(i,0);
			j = sCurrent+i;
			k = 0;
			if ( nStart != -1 ) {
			    if	    ( (j>=nStart)&&(j<=nEnd) ) { k = 1;}
			    else if ( (j<=nStart)&&(j>=nEnd) ) { k = 1;}
			}
			wnd_puts(prints(" %-80s",szTextBuf[j]),k?0x1e:0xf0);
		    }
		    wnd_select(widMain);
		    if ( nStart == -1 ) {
			wnd_ftext(bStart,"    ",0x2f0);
			wnd_ftext(bEnd,"    ",0x2f0);
		    }
		    else {
			wnd_ftext(bStart,prints("%4d",nStart),0x2f0);
			wnd_ftext(bEnd,prints("%4d",nEnd),0x2f0);
		    }
		}
		sUpd = 0;
	    }

	    which = wnd_gbtn(0);
	    if ( (which >= bSBarFirst) &&
		 (which <= bSBarLast) ) {
		j = sCurrent+which-bSBarFirst;
		if ( j < sTotal ) {
		    if ( nStart == -1 ) {
			nStart = nEnd = j;
			if ( nHasStateFnc ) {
			    wnd_state(bUnMark,1,0);
			    wnd_state(bOK,1,0);
			}
		    }
		    else
			nEnd = j;
		    sUpd=2;
		}
		continue;
	    }

	    switch(which) {
	      case bSBar:	 // Up arrow
		if ( sCurrent > 0 ) {
		    sCurrent--;
		    wnd_select(widSub);
		    wnd_rolldn(rSub,1);
		    wnd_loct(0,0);
		    j = sCurrent;
		    k = 0;
		    if ( nStart != -1 ) {
			if ( (j>=nStart)&&(j<=nEnd) )
			    k = 1;
			else if ( (j<=nStart)&&(j>=nEnd) )
			    k = 1;
		    }
		    wnd_puts(prints(" %-80s",szTextBuf[j]),k?0x1e:0xf0);
		}
		sUpd++;
		continue;
	      case bSBar+1:	 // Down arrow
		if ( (sCurrent+sPageSize)>=sTotal )
			continue;
		if ( sCurrent < sTotal ) {
		    sCurrent++;
		    wnd_select(widSub);
		    wnd_rollup(rSub,1);
		    wnd_loct(sPageSize-1,0);
		    j = (sCurrent+sPageSize-1);
		    k = 0;
		    if ( nStart != -1 ) {
			if ( (j>=nStart)&&(j<=nEnd) )
			    k = 1;
			else if ( (j<=nStart)&&(j>=nEnd) )
			    k = 1;
		    }
		    wnd_puts(prints(" %-80s",szTextBuf[j]),k?0x1e:0xf0);
		}
		sUpd++;
		continue;
	      case bSBar+2:	 // Activation
		which = wnd_vscroll(2,bSBar);
		switch(which) {
		  case -2:	// Upper activation
		    if ( sCurrent == 0 )
			break;
		    sUpd=2;
		    sCurrent -= sPageSize;
		    if ( sCurrent < 0 )
			sCurrent = 0;
		    break;
		  case -1:	// Lower activation
		    if ( (sCurrent+sPageSize)>=sTotal )
			break;
		    sCurrent += sPageSize;
		    if ( sCurrent > (sTotal-sPageSize) )
			sCurrent = sTotal-sPageSize;
		    if ( sCurrent <= 0 )
			sCurrent = 0;
		    sUpd=2;
		    break;
		  default:
		    which--;
		    if ( which > (sTotal-sPageSize) )
			which = sTotal-sPageSize;
		    if ( which <= 0 )
			which = 0;
		    if ( sCurrent == which )
			break;
		    sCurrent = which;
		    sUpd=2;
		}
		continue;
	      case bOK:
		wndclose();
		return 1;
	      case -1:
	      case 0:
	      case bCancel:
		wndclose();
		return 0;
	      case -2:
	      case -3:
	      case bUnMark:
		nStart = nEnd = -1;
		if ( nHasStateFnc ) {
		    wnd_state(bUnMark,1,1);
		    wnd_state(bOK,1,1);
		}
		sUpd=2;
		continue;
	      case bMarkAll:
		nStart = 0;
		nEnd = sTotal-1;
		if ( nHasStateFnc ) {
		    wnd_state(bUnMark,1,0);
		    wnd_state(bOK,1,0);
		}
		sUpd=2;
		continue;
	      case bStart:
		nStart = wnd_fgetd(bStart,4,0x2e0,nStart);
		if ( nEnd<0 || nStart<0 )
			nEnd = nStart;
		if ( nHasStateFnc ) {
		    j = (nStart<0)?1:0;
		    wnd_state(bUnMark,1,j);
		    wnd_state(bOK,1,j);
		}
		sUpd=2;
		continue;
	      case bEnd:
		nEnd = wnd_fgetd(bEnd,4,0x2e0,nEnd);
		if ( nEnd<0 || nStart<0 )
			nStart = nEnd;
		if ( nHasStateFnc ) {
		    j = (nStart<0)?1:0;
		    wnd_state(bUnMark,1,j);
		    wnd_state(bOK,1,j);
		}
		sUpd=2;
		continue;
	      default:
		continue;
	    }
	    break;
	}
	wndclose();
	return 0;
}
//
//	Some useful routine supports
//
int	nSupport=identify("wnd_resize");
int	WndButton(int r,int c, int bsize, char text[])
{
	int	i;
	if ( nSupport ) 	// Having direct support ?
		return wnd_button(r,c,bsize,0x1277,text);
	wnd_loct(r,c);
	i = wnd_field(0,0x1277,bsize);
	wnd_puts(text,0x270);
	return i;
}
//
//	WndGroup() -- Create Groupping window effect
//
void	WndGroup(int r,int c,int width, int height, char text[], int color)
{
	wnd_loct(r,c);
	wnd_rloct(height,0);
	if ( IsWin )
		wnd_field(0,0x277,(height<<8)|width);
	else
		wnd_field(0,((color&0xff)^0x08)|0x200,(height<<8)|width);
	wnd_itemno(1,-1);
	wnd_rloct(-height,2);
	if ( !IsWin )
		wnd_rloct(0xfc00,0);
	put3Dstr(text,color);
	wnd_loct(r,c);	      // Reset row offset, fix DOS version bug
}
//
//	put3Dstr() -- Drawing 3D text
//
void	put3Dstr(char str[],int atrb)
{
	int	nLen;
	if ( nSupport ) {
		wnd_puts(str,atrb|0x300);
		return;
	}
	nLen=len(str);
	if ( nLen ) {
		wnd_rloct(0x100,IsWin?0x100:0);
		wnd_puts(str,(atrb^0x0f));
		wnd_rloct(0xff00,-nLen);
		wnd_puts(str,(atrb&0xff0f)|0x8000);
	}
}
//
#define OBM_COMBO	"#32738"
int	WndEntry(int row, int col, char str[],int fsize,int combo)
{
	int	bID;

	wnd_loct(row,col);
	put3Dstr(str,nSatrb);
	bID = wnd_field(0,IsWin?0x1f0:0x21f0,fsize);
	if ( combo ) {
		wnd_rloct(0xfe00,0x0600|(fsize&0xff));
		if ( IsWin) wnd_runit(12);
		if ( identify("cb_empty") ) {   // Supported only after this version
		     wnd_icon(1,0xA77,0x101,OBM_COMBO);
		     wnd_runit(-1);
		     wnd_rloct(0x0200,0x0002);	// locate back
		}
		else {
		     wnd_field(0,IsWin?0x277:0x370,0x101);
		     wnd_runit(IsWin?-1:16);
		     wnd_ftext(bID+1,"",IsWin?0x277:0x270);
		     wnd_rloct(0x0200,0x0201);	// locate back
		}
	}
	return bID;
}

char	szAsciiText[][70]={	    // Default English Display Text
/*  0 */ "\nNo text available from clipboard.",
/*  1 */ "\nNo effective text available from clipboard.",
/*  2 */ "\nText Insert Point: ",
/*  3 */ "\nText Insert Angle <0.>: ",
/*  4 */ "Insert Texts from Clipboard [%d lines read]",
/*  5 */ "Please Mark the Text to Insert",
/*  6 */ "&OK",
/*  7 */ "&Cancel",
/*  8 */ "&All",
/*  9 */ "&UnSelect",
/* 10 */ "Range: ",
};

char	szBig5Text[][70]={	   // Default Chinese Big-5 Display Text
/*  0 */ "\nŶKïSrA.",
/*  1 */ "\nŶKïrAƵLĤr.",
/*  2 */ "\nЫwrJI: ",
/*  3 */ "\nЫwrJ <0.>: ",
/*  4 */ "ѰŶKïJre %d ŪJf",
/*  5 */ "г]wr檺d",
/*  6 */ "Tw[&O]",
/*  7 */ "[&C]",
/*  8 */ "[&A]",
/*  9 */ "[&U]",
/* 10 */ "rd: ",
};

char	szGB2312Text[][70]={	// Default Chinese GB2312 Display Text
/*  0 */ "\nû̬.",
/*  1 */ "\n̬Ч.",
/*  2 */ "\nֲָ: ",
/*  3 */ "\nֲָǶ <0.>: ",
/*  4 */ "ɼڲ֣ %d ж",
/*  5 */ "趨еķΧ",
/*  6 */ "ȷ[&O]",
/*  7 */ "ȡ[&C]",
/*  8 */ "ȫѡ[&A]",
/*  9 */ "ѡ[&U]",
/* 10 */ "зΧ: ",
};
//
//	The following Japenese messages are translated by _____________.
//
char	szSJISText[][70]={    // Default Japenese SJIS Display Text
/*  0 */ "\nNbv{[hɃeLXg܂B",
/*  1 */ "\nNbv{[hɗLȃeLXg܂B",
/*  2 */ "\neLXg̑}_: ",
/*  3 */ "\npx <0.>: ",
/*  4 */ "Nbv{[heLXg}܂B[%d sǍ] ",
/*  5 */ "}eLXg}[NĉB",
/*  6 */ "&OK",
/*  7 */ "ݾ[&C]",
/*  8 */ "S[&A]",
/*  9 */ "I",
/* 10 */ "͈: ",
};
//
//	The following Korean messages are translated by _____________.
//
char	szKoreanText[][70]={	// Default Korean Display Text
/*  0 */ "\nNo text available from clipboard.",
/*  1 */ "\nNo effective text available from clipboard.",
/*  2 */ "\nText Insert Point: ",
/*  3 */ "\nText Insert Angle <0.>: ",
/*  4 */ "Insert Texts from Clipboard [%d lines read]",
/*  5 */ "Please Mark the Text to Insert",
/*  6 */ "&OK",
/*  7 */ "&Cancel",
/*  8 */ "&All",
/*  9 */ "&UnSelect",
/* 10 */ "Range: ",
};

#define BIG5		0x404
#define GB2312		0x804
#define SJIS		0x411
#define KOREAN		0x412
#define ASCII		0x409

int	SetMsgText()
{
	char	szBuf[80];
	int	switched;

	switch(1) {  // Test if we are in TwinCAD ...
	  case !identify("os_type"):
	  case (os_type()<0):	// Must be in DOS version!
		return 1;
	}

	szBuf = get_profile("General Setup","AlternateLanguage");
	switched = 0;
	if ( szBuf!="" )
	    switched = eval(szBuf);
	if ( switched ) {
	    switch(os_language()) {
	      case BIG5:    setdata(szText,szBig5Text);     break;
	      case GB2312:  setdata(szText,szGB2312Text);   break;
	      case SJIS:    setdata(szText,szSJISText);     break;
	      case KOREAN:  setdata(szText,szKoreanText);   break;
	      default:	    setdata(szText,szAsciiText);    break;
	    }
	}
	else
	    setdata(szText,szAsciiText);
	return 0;
}
